/*
 * milena_phrasi.c - Milena TTS system (phraser)
 * Copyright (C) Bohdan R. Rau 2008 <ethanak@polip.com>
 *
 * Milena is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 *
 * Milena is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with Milena.  If not, write to:
 * 	The Free Software Foundation, Inc.,
 * 	51 Franklin Street, Fifth Floor
 * 	Boston, MA  02110-1301, USA.
 */

static char lci[256]={
	0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
	0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
	0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
	'0','1','2','3','4','5','6','7','8','9',0,0,0,0,0,0,
	0,'a','b','c','d','e','f','g','h','i','j','k','l','m','n','o',
	'p','q','r','s','t','u','v','w','x','y','z',0,0,0,0,0,
	0,'a','b','c','d','e','f','g','h','i','j','k','l','m','n','o',
	'p','q','r','s','t','u','v','w','x','y','z',0,0,0,0,0,
	0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
	0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
	0,'',0,'',0,'','',0,0,'','','','',0,'','',
	0,'',0,'',0,'','',0,0,'','','','',0,'','',
	'','','','','','','','','','','','','','','','',
	'','','','','','','',0,'','','','','','','','',
	'','','','','','','','','','','','','','','','',
	'','','','','','','',0,'','','','','','','',0};

static char uci[256]={
	0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
	0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
	0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
	0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
	0,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,
	1,1,1,1,1,1,1,1,1,1,1,0,0,0,0,0,
	0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
	0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
	0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
	0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
	0,1,0,1,0,1,1,0,0,1,1,1,1,0,1,1,
	0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
	1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,
	1,1,1,1,1,1,1,0,1,1,1,1,1,1,1,0,
	0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
	0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0};

static char *aps[]={
	"em","u","w","om","ach","y",
	"ego","emu","a","owi","ym","ie",
	"owy","owego","owemu",
	"owej","owa","owe","owym","owych",
	"owymi","owie","ami","owi",
	"owski","owskiego","owskiemu",
	"owskim","owska","owsk","owskiej",
	"owscy","owskie","owskich","owskimi","s",NULL};


static int good_char(struct milena *cfg,char c)
{
	if (isdigit(c)) return 1;
	c=lci[c & 255];
	if (!c) return 0;
	return cfg->letters[c&255] & 3;
}

#define MLWORD_SPEAKABLE 1
#define MLWORD_NUMBER 2
#define MLWORD_UNSPEAK 3

/* MLWORD_NICKY - np. 'dupa23' */
#define MLWORD_NICKY 4


static char *cyfra[]={"zero","jeden","dwa","trzy","cztery",
		"pi","sze","siedem","osiem",
		"dziewi","dziesi","jedenacie",
		"dwanacie","trzynacie","czternacie",
		"pitnacie","szesnacie","siedemnacie",
		"osiemnacie","dziewitnacie"};
static char *dzies[]={"","","dwadziecia","trzydzieci",
		"czterdzieci","pidziesit","szedziesit",
		"siedemdziesit","osiemdziesit","dziewidziesit"};
static char *setka[]={"zero","sto","dwiecie","trzysta","czterysta",
		"piset","szeset","siedemset","osiemset","dziewiset"};

static char *dits_tys[]={"tysic","tysice","tysicy"};
static char *dits_mln[]={"milion","miliony","milionw"};

static char *milena_get_unit(struct milena *cfg,char **str,int last_num,int *female,int hidden);

static int what_dit(int dit)
{
	if (dit==1) return 0;
	dit%=100;
	if (!dit) return 2;
	if (dit>19) dit%=10;
	if (dit>=2 && dit<=4) return 1;
	return 2;
}

static int milena_integer(int nr,char *outbuf,int buflen,int pos)
{

	int mln,tys,blank=0;
	void numeras(int n)
	{
		int setki=n/100,dz;
		n=n%100;
		if (setki) {
			if (blank) pushbuf(' ');
			blank=1;
			pushstr(setka[setki]);
		}
		if (!n) return;
		if (blank) pushbuf(' ');
		blank=1;
		if (n<20) {
			pushstr(cyfra[n]);
			return;
		}
		dz=n/10;
		n%=10;
		pushstr(dzies[dz]);
		if (!n) return;
		pushbuf(' ');
		pushstr(cyfra[n]);
	}

	mln=nr/1000000;
	nr=nr%1000000;


	if (mln) {
		if (mln) numeras(mln);
		if (blank) pushbuf(' ');blank=1;
		pushstr(dits_mln[what_dit(mln)]);
	}
	tys=nr/1000;
	nr=nr%1000;
	if (tys) {
		if(tys > 1 || mln) numeras(tys);
		if (blank) pushbuf(' ');blank=1;
		pushstr(dits_tys[what_dit(tys)]);
	}
	if (nr) numeras(nr);
	return pos;
}

static int milena_digitize(char **str,char *outbuf,int buflen,int pos,int spell,int *outmode,struct milena *cfg)
{
	int blank=0;
	int ndit;
	int mln,tys,nr;
	static char *dvs[]={"","","drug","trzeci","czwart","pit",
		"szst","sidm","sm","dziewit","dziesit",
		"jedenast","dwunast","trzynast","czternast",
		"pitnast","szesnast"};

	void numerek(int n,int dits,int fema)
	{
		int setki=n/100,dz;
		if (!n && dits==1) {
			if (blank) pushbuf(' ');
			pushstr("zero");
			return;
		}
		n=n%100;
		if (!setki && dits>2) {
			if (blank) pushbuf(' ');
			pushstr("zero");
			if (n<10) {
				pushstr(" zero ");
				if (fema && n==2) {
					pushstr("dwie");
				}
				else {
					pushstr(cyfra[n]);
				}
				return;
			}
		}
		if (setki) {
			if (blank) pushbuf(' ');
			blank=1;
			pushstr(setka[setki]);
		}
		if (n<10 && !setki && dits==2) {
			if (blank) pushbuf(' ');
			blank=1;
			pushstr("zero ");
			if (fema && n==2) {
				pushstr("dwie");
			}
			else {
				pushstr(cyfra[n]);
			}
			return;
		}
		if (!n) return;
		if (blank) pushbuf(' ');
		blank=1;
		if (n<20) {
			if (fema && n==2) {
				pushstr("dwie");
			}
			else {
				pushstr(cyfra[n]);
			}
			return;
		}
		dz=n/10;
		n%=10;
		pushstr(dzies[dz]);
		if (!n) return;
		pushbuf(' ');
		if (fema && n==2) {
			pushstr("dwie");
		}
		else {
			pushstr(cyfra[n]);
		}
	}


	int get_num(int ile)
	{
		int nr=0;
		while (ile>0) {
			ile--;
			nr=nr*10+ (*(*str)++) -'0';
		}
		return nr;
	}

	int is_frac(char *s)
	{
		int n;
		if (*s!='.' && *s!=',') return 0;
		s++;
		if (!*s || !isdigit(*s)) return 0;
		for (n=0;*s && isdigit(*s); s++,n++);
		if ((*s=='.' || *s==',') && s[1] && isdigit(s[1])) return 0;
		return n;
	}

	void digitize_frac(int nd)
	{

		(*str)++;
		pushstr(" przecinek");
		blank=1;
		if (nd<=3) {
			int n=get_num(nd);
			numerek(n,nd,0);
			return;
		}
		while (**str && isdigit(**str)) {
			int n=*(*str)++;
			pushbuf(' ');
			pushstr(cyfra[n&15]);
		}
	}
	int vulgar(void)
	{
		char *s=*str;
		int n1=strtol(s,&s,10),n2;
		if (*s++!='/') return 0;
		if (*s<'1' || *s>'9') return 0;
		n2=strtol(s,&s,10);
		if (n2<2 || (n2>10 && n2 != 16)) return 0;
		if (n1>=2*n2) return 0;
		if (*s && !isspace(*s)) {
			if (lci[(*s) & 255]) return 0;
			if (strchr("/.-",*s)) {
				if (lci[s[1] & 255]) return 0;
			}
		}
		*str=s;
		if (n1==1) {
			pushstr("jedna");
		}
		else {
			numerek(n1,0,1);
		}
		pushbuf(' ');
		pushstr(dvs[n2]);
		if (n1==1) {
			pushbuf('a');
			return 1;
		}
		if ((n1>=2 && n1<=4) || (n1 >=22 && n1<=24) || n1==32) {
			if (n2==2) {
				pushstr("ie");
			}
			else {
				pushbuf('e');
			}
			return 1;
		}
		if (n2==2) {
			pushstr("ich");
		}
		else {
			pushstr("ych");
		}
		return 1;
	}

	int is_ip_part(char **s)
	{
		int n;
		if (!isdigit(**s)) return -1;
		if (**s=='0') {
			(*s)++;
			if (!**s || !isdigit(**s)) return 0;
			return -1;
		}
		n=strtol(*s,s,10);
		if (n<256) return n;
		return -1;
	}

	void push_dot(int n1,int n2)
	{
		if (!n1 || !n2) return;
		if ((n1%100 == 0 && n2<100) ||
			(n2<100 && n1%10==0 && n1!=10 && n1!=110 && n1!=210)) {
			if (blank) pushbuf(' ');
			blank=1;
			pushstr("kropka");
		}
	}
	int is_ip(void)
	{
		char *s;
		int n1,n2,n3,n4;
		s=*str;
		n1=is_ip_part(&s);
		if (n1<0) return 0;
		if (*s++!='.') return 0;
		n2=is_ip_part(&s);
		if (n2<0) return 0;
		if (*s++!='.') return 0;
		n3=is_ip_part(&s);
		if (n3<0) return 0;
		if (*s++!='.') return 0;
		n4=is_ip_part(&s);
		if (n4<0) return 0;
		if (lci[(*s)&255]) return 0;
		if (*s=='.' && isdigit(s[1])) return 0;
		if (n1) numerek(n1,1,0);else {
			if(blank) pushbuf(' ');
			blank=1;
			pushstr("zero");
		}
		push_dot(n1,n2);
		if (n2) numerek(n2,1,0);
		else {
			if(blank) pushbuf(' ');
			blank=1;
			pushstr("zero");
		}
		push_dot(n2,n3);
		if (n3) numerek(n3,1,0);
		else {
			if(blank) pushbuf(' ');
			blank=1;
			pushstr("zero");
		}
		push_dot(n3,n4);
		if (n4) numerek(n4,1,0);
		else {
			if(blank) pushbuf(' ');
			blank=1;
			pushstr("zero");
		}
		*str=s;
		return 1;
	}
	int is_female_unit()
	{
		char *s=*str;
		int female=0;
		while (*s && isdigit(*s)) s++;
		if (*s=='.' && s[1] && isdigit(s[1])) {
			s+=2;
			while (*s && isdigit(*s)) s++;
		}
		if (!isspace(*s)) return 0;
		while (*s && isspace(*s)) s++;
		s=milena_get_unit(cfg,&s,0,&female,0);
		if (!s) return 0;
		return female;
	}

	int female=is_female_unit();
	if (!spell && is_ip()) {
		if (outmode) *outmode=-1;
		return pos;
	}

	while (**str=='0') {
		if (blank) pushbuf(' ');
		blank=1;
		pushstr("zero");
		(*str)++;
		if (**str!='.' && **str!=',') spell=1;
	}
	if (!spell) {
		if (vulgar()) {
			if (outmode) *outmode=3;
			return pos;
		}
	}

	for (ndit=0;(*str)[ndit] && isdigit((*str)[ndit]);ndit++);
	if (ndit>9) spell=1;
	if (outmode) *outmode=2;
	if (spell) {
		int sout;
		if (ndit<4) {
			numerek(sout=get_num(ndit),ndit,0);
			if (outmode) *outmode=what_dit(sout);
			return pos;
		}
		if (ndit==4) {
			numerek(get_num(2),2,0);
			numerek(get_num(2),2,0);
			return pos;
		}
		while(ndit>3) {
			numerek(get_num(2),2,0);
			ndit-=2;
		}
		numerek(get_num(ndit),ndit,0);
		return pos;
	}
	if (ndit) {
		nr=get_num(ndit);
		if (outmode) *outmode=what_dit(nr);
		if (female && nr == 1) {
			if (blank) pushbuf(' ');blank=1;
			pushstr("jedna");
		}
		else {
			mln=nr/1000000;
			nr=nr%1000000;
			if (mln) {
				if (mln) numerek(mln,0,0);
				if (blank) pushbuf(' ');blank=1;
				pushstr(dits_mln[what_dit(mln)]);
			}
			tys=nr/1000;
			nr=nr%1000;
			if (tys) {
				if(tys > 1 || mln) numerek(tys,0,0);
				if (blank) pushbuf(' ');blank=1;
				pushstr(dits_tys[what_dit(tys)]);
			}
			if (nr) numerek(nr,0,female);
		}
	}
	if (!spell) {
		int nd;
		if ((nd=is_frac(*str))) {
			digitize_frac(nd);
			if (outmode) *outmode=3;
		}
	}
	return pos;
}


static char *milena_pseudo(struct milena *cfg,char **str)
{
	struct milena_pseudo *ps;

	char *compare_pseudo(char *s1,char *pat)
	{
		char dn;
		while (*pat) {
			char cn=lci[(*s1) & 255];
			if (cn != *pat) return NULL;
			pat++;
			s1++;
			if (*pat=='*') {
				pat++;
				while (lci[(*s1) & 255] == cn) s1++;
			}
		}
		if (!*s1) return s1;
		dn=lci[(*s1) & 255];
		if (!dn) return s1;
		if (isdigit(dn)) return NULL;
		if (cfg->letters[dn & 255] & 3) return NULL;
		return s1;
	}


	for (ps=cfg->pseudowords;ps;ps=ps->next) {
		char *s=compare_pseudo(*str,ps->pattern);
		if (s) {
			*str=s;
			return ps->result;
		}
	}
	return NULL;
}

static char *milena_get_unit(struct milena *cfg,char **str,int last_num,int *female,int hidden)
{
	struct milena_nflex *nf;

	char *compare_nflex(char *s1,char *pat)
	{
		char dn;
		while (*pat) {
			if (*s1 != *pat) return NULL;
			pat++;
			s1++;
		}
		if (!*s1) return s1;
		dn=lci[(*s1) & 255];
		if (!dn) return s1;
		if (isdigit(dn)) return NULL;
		if (cfg->letters[dn & 255] & 3) return NULL;
		return s1;
	}
	for (nf=cfg->nflex;nf;nf=nf->next) {
		if (!hidden && (nf->flags & MILENA_NFLEX_HIDDEN)) continue;
		char *s=compare_nflex(*str,nf->pattern);
		if (s) {
			*str=s;
			if (female) *female=nf->flags & MILENA_NFLEX_FEMALE ;
			return nf->result[last_num];
		}
	}
	return NULL;

}


static int milena_spell(struct milena *cfg,char **str,char *outbuf,int buflen,int pos,int last_num)
{
	int blank=0,lws=0,flet=0;
	char *p;
	char *spellsa[]={
		"a","be","c~'e","de","e","ef","gie","ha","i","jot",
		"ka","el","em","en","o","pe","ku","er","es","te",
		"u","fa","wu","iks","igrek","zet"};
	char *spellsb[]={"e","et","zi","si","ci","ni",NULL};
	char *find_spell(char c)
	{
		int i;
		if (c>='a' && c<='z') return spellsa[c-'a'];
		for (i=0;spellsb[i];i++) if(spellsb[i][0]==c) return spellsb[i]+1;
		return NULL;
	}


	if ((p=milena_pseudo(cfg,str))) {
		pushstr(p);
		return pos;
	}
	for (;**str;) {
		char cx,*cy,*c;
		int i,stres;
		if (isdigit(**str)) {
			if (blank) pushbuf(' ');
			pos=milena_digitize(str,outbuf,buflen,pos,1,NULL,cfg);
			blank=1;lws=flet=0;
			continue;
		}
		cx=lci[(**str) & 255];
		if (!cx) break;
		(*str)++;
		if (find_spell(cx)) {
			stres=1;
			if (cx == 'y') stres=2;
			for (i=0;(*str)[i];i++) {
				char cz=lci[(*str)[i] & 255];
				if (cz=='y') stres=2;
				else if (find_spell(cz)) stres=1;
				else break;
			}
			if (blank) pushbuf(' ');
			if (stres==1) {
				pushstr("[1]");
			}
			else {
				pushstr("[2]");
			}
			lws=0;
			cy=find_spell(cx);
			for(;;) {
				if (lws) {
					if (((lws=='s' || lws=='r' || lws=='t' || lws == 'k') && (*cy=='z' || *cy=='h')) ||
					    (strchr("ns",lws) && *cy=='i')) {
					    	pushstr("~'");
					}
					else if (strchr("aeiour",lws) && strchr("aeiou",*cy)) {
						pushbuf('_');
					}
					else if (lws == 't' && *cy=='t') {
						pushstr("~'");
					}
					else if (lws == 'n' && *cy=='n') {
						pushstr("_~'");
					}
				}
				pushstr(cy);
				while (cy[1]) cy++;
				lws=*cy;
				cx=lci[(**str) & 255];
				cy=find_spell(cx);
				if (!cy) break;
				(*str)++;
			}
			blank=1;
			continue;
		}
		if(blank) pushbuf(' ');
		if (cx=='') c="~!a z ogonkiem";
		else if (cx=='') c="~!e z ogonkiem";
		else if (cx=='') c="~!o z kresk";
		else c="litera";
		pushstr(c);
		blank=1;
	}
	return pos;
}

struct phraser_dic_rc {
	char *str;
	int len;
};

static int phraser_CompareDic(struct milena *cfg,char *patr,char **ostr,struct phraser_dic_rc *array,int *onph)
{
	char a_char(char t)
	{
		char tc;
		tc=lci[t & 255];
		if (tc) return tc;
		return t;
	}

	int compare_char(char pchar,char schar)
	{
		if (pchar == schar) return 1;
		if (uci[schar & 255]) return 0;
		pchar=lci[pchar & 255];
		if (pchar == schar) return 1;
		return 0;
	}

	char *str=*ostr;
	int nphr=0;
	while (*patr) {
		char cpat=*patr++;
		if (*str=='~' && !(cfg->phraser_mode & MILENA_PHR_IGNORE_TILDE) && strchr(ESCHAR,str[1])) return 0;
		if (cpat=='\\') {
			if (*patr++ == *str) str++;
			continue;
		}
		if (cpat=='\'') {
			if (*str!='\'' && *str != '`') return 0;
			str++;
			while (*str && isspace(*str)) str++;
			continue;
		}
		if (cpat=='_') { /* opcjonalna spacja */
			while (*str && isspace(*str)) str++;
			continue;
		}
		if (cpat=='+') { /* spacja */
			if (!*str || !isspace(*str)) return 0;
			str++;
			while (*str && isspace(*str)) str++;
			continue;
		}
		if (cpat=='`') {
			if (isspace(*str)) {
				char *ss=str+1;
				while(isspace(*ss)) ss++;
				if (*ss=='\'' || *ss=='`') {
					ss++;
					while (*ss && isspace(*ss)) ss++;
					str=ss;
					continue;
				}
				return 0;
			}
			if (*str=='\'' || *str=='`') {
				str++;
				while (*str && isspace(*str)) str++;
			}
			continue;
		}
		if (cpat=='~') {
			if (*patr == '~') { // dowolny cig spacji i mylnikw
				patr++;
				while (*str && (isspace(*str) || *str=='-')) str++;
				continue;
			}
			if (*str=='-') {
				str++;
				continue;
			}
			//while (*str && isspace(*str)) str++;
			//if (*str!='-') continue;
			//str++;
			while (*str && isspace(*str)) str++;
			continue;
		}
		if (cpat=='[') {
			int found=0;
			while (*patr) {
				cpat=*patr++;
				if (cpat==']') break;
				if (!found && compare_char(*str,cpat)) found=1;
			}
			if (!found) return 0;
			str++;
			continue;
		}
		if (cpat=='*') {
			int i;
			if (nphr >= 10) return 0;
			array[nphr].str=str;
			for (i=0;str && lci[(*str) & 255];str++,i++);
			array[nphr++].len=i;
			continue;
		}
		if (cpat=='(') {
			char *bstr;
			int i;
			if (nphr>=10) return 0;
			while (*patr) {
				bstr=patr;
				for (i=0;bstr[i];i++) {
					if (bstr[i]=='|' || bstr[i]==')') break;
					if (!compare_char(str[i],bstr[i])) break;
				}
				if (!bstr[i]) return 0;
				if (bstr[i]==')' || bstr[i]=='|') break;
				for (;*patr;patr++) {
					if (*patr==')') return 0;
					if (*patr=='|') {
						patr++;
						break;
					}
				}
				bstr=patr;
			}
			str+=i;
			array[nphr].len=-1;
			array[nphr++].str=bstr;
			patr=strchr(patr,')');
			if (!patr) return 0;
			patr++;
			continue;
		}
		if (!compare_char(*str,cpat)) return 0;
		str++;
	}
	if (*str) {
		char *c=str-1;
		/* wntrze */
		if (*str=='~' && !(cfg->phraser_mode & MILENA_PHR_IGNORE_TILDE) && strchr(ESCHAR,str[1])) return 0;

		/* apos? */
		if (lci[(*c) & 255] && *str=='\'' && lci[str[1]&255]) {
			int i,j;
			for (i=0;aps[i];i++) {
				for (j=0;aps[i][j];j++) if (aps[i][j]!=lci[str[j+1]&255]) break;
				if (aps[i][j]) continue;
				if (!lci[str[j+1] & 255]) return 0;
			}
		}
		if (*c=='.' && *str=='.' && str[1]=='.') return 0;
		if (lci[(*str)&255]) return 0;
	}
	*ostr=str;
	*onph=nphr;
	return 1;

}

static int phraser_OutputDic(struct milena *cfg,
	struct milena_udic *ud,
	int pos,
	char *outbuf,
	int buflen,
	struct phraser_dic_rc *array,
	int narr)
{
	int npr=0;
	char *c;
	if (!ud->pron || !strpbrk(ud->pron,"\t ")) {
		if (ud->flags & MILDIC_V) {
			pushstr("{v}");
		}
		if ((ud->flags & ~ MILDIC_V) || ud->stress || ud->pstress) {
			pushbuf('[');
			if (ud->stress) pushbuf('0'+ud->stress);
			if (ud->pstress) {
				pushbuf('+');
				pushbuf('0'+ud->pstress);
			}
			if (ud->flags & MILDIC_K) pushbuf('k');
			if (ud->flags & MILDIC_U) pushbuf('n');
			pushbuf(']');
		}
	}
	if (!ud->pron) c=ud->word;
	else c=ud->pron;
	for (;*c;c++) {
		char d;
		if (*c!='%') {
			d=lci[(*c) & 255];
			if (d) pushbuf(d);
			else pushbuf(*c);
			continue;
		}
		if (isdigit(c[1])) {
			c++;
			npr=(*c) - '0';
		}
		if (npr<narr) {
			char *s,i;
			i=array[npr].len;
			s=array[npr++].str;
			if (i<0) {
				for (;*s && *s!=')' && *s!='|';s++) {
					d=lci[(*s) & 255];
					if (d) pushbuf(d);
					else pushbuf(*s);
				}
			}
			else {
				for (;*s && i>0;i--,s++) {
					d=lci[(*s) & 255];
					if (d) pushbuf(d);
					else break;
				}
			}
		}
	}
	return pos;
}

static int milena_nicky(struct milena *cfg,char **str,char *outbuf,int buflen,int pos,int last_num)
{
	char nick[64];
	char *si,*so;int n,done;
	struct milena_udic *ud;
	for (si=*str,n=0;good_char(cfg,*si);n++,si++) {
		if (isdigit(*si)) break;
		if (n>=63) return milena_spell(cfg,str,outbuf,buflen,pos,last_num);
		nick[n]=*si;
	}
	nick[n]=0;
	*str=si;
	if (pos) {
		pushbuf(' ');
	}
	for (done=0,ud=cfg->udic;ud;ud=ud->next) {
		int nphr;
		struct phraser_dic_rc phra[16];
		char *cs;
		if (ud->flags & MILDIC_ATSTART) {
			if(pos) continue;
		}
		if (ud->flags & MILDIC_ATQUE) continue;
		if (ud->flags & MILDIC_BEFORECAP) continue;
		if (ud->flags & MILDIC_ATEND) continue;
		cs=nick;
		if (!phraser_CompareDic(cfg,ud->word,&cs,phra,&nphr)) continue;
		if (ud -> flags & MILDIC_SPELL) {
			return milena_spell(cfg,str,outbuf,buflen,pos,last_num);
		}
		pos=phraser_OutputDic(cfg,ud,pos,outbuf,buflen,phra,nphr);
		done=1;
		break;
	}
	if (!done) pushstr(nick);
	pushbuf(' ');
	if (!**str) return pos;
	return milena_spell(cfg,str,outbuf,buflen,pos,last_num);
}


char *mi_spell_all(char c)
{
	static char *spl[]={
		"\x80wielokropek",
		".kropka",
		",przecinek",
		":dwukropek",
		"(nawias",
		")prawy nawias",
		"\"cudzysw",
		"'apostrof",
		";rednik",
		"-mylnik",
		"[nawias kwadratowy",
		"]prawy nawias kwadratowy",
		"{klamra",
		"}prawa klamra",
		"?pytajnik",
		"!wykrzyknik",NULL};
	int i;
	for (i=0;spl[i];i++) if (spl[i][0]==c) return spl[i]+1;
	return NULL;
}

char *mi_spell(struct milena *cfg,char c,int punct,char *some)
{
	struct milena_spell *ms;
	for (ms=cfg->spellers;ms;ms=ms->next) if (ms->znak == c) return ms->result;
	if ((punct == 1 && some && strchr(some,c)) || punct == 2) {
		return mi_spell_all(c);
	}
	return NULL;
}

static char *phraser_get_emot(struct milena *cfg,char *str,char **outstr,int *eof)
{
	char *c,*d;
	struct milena_emot *me;
	static char emotbuf[64];
	int fin=0;
	if (!cfg->emots) return NULL;
	if (eof) *eof=0;
	for (me=cfg->emots;me;me=me->next) {
		for (c=str,d=me->emot;*d;) {
			if (*d=='+') {
				if (*c++!=' ') break;
				d++;
				continue;
			}
			if (*d=='\\') {
				d++;
				if (!*d) break;
			}
			if (*c++ != *d) break;
			d++;
		}
		if (!*d && strlen(me->result)<58) {
			if (outstr) *outstr=c;
			d=emotbuf;
			if (cfg->emotmod) {
				sprintf(d,"{%d}",cfg->emotmod);
				d+=strlen(d);
			}
			strcpy(d,me->result);
			if (strchr("?!",me->result[strlen(me->result-1)])) {
				fin=1;
				if (emotbuf[strlen(emotbuf)-1]=='?') fin=4;
				emotbuf[strlen(emotbuf)-1]=0;
				if (eof) *eof=fin;
			}
			if (cfg->emotmod) strcat(emotbuf,"{0}");
			return emotbuf;
		}
	}
	c=str;
	if (*c=='<') {
		d=emotbuf;
		if (cfg->emotmod) {
			sprintf(d,"{%d}",cfg->emotmod);
			d+=strlen(d);
		}
		c++;
		while (*c && !isdigit(*c) && lci[(*c) & 255]) {
			if (d-emotbuf > 58) break;
			*d++=*c++;
		}
		if (*c=='!' || *c=='?') {
			fin=1;
			if (*c=='?') fin=4;
			*c++;
		}
		if (*c=='>') {
			if (eof) *eof=fin;
			if (cfg->emotmod) strcpy(d,"{0}");
			else *d=0;
			c++;
			if (outstr) *outstr=c;
			return emotbuf;
		}
	}
	return NULL;
}

int milena_GetParType(struct milena *cfg,char **str,int spell_empty)
{
	int dialog=1;
	char *c,*d;
	c=NULL;
	for (;**str;(*str)++) {
		if (phraser_get_emot(cfg,*str,NULL,NULL)) return dialog;
		//struct milena_emot *me;
		//for (me=cfg->emots;me;me=me->next) if (!strncmp(me->emot,*str,strlen(me->emot))) {
			//return dialog;
		//}
		if (!(cfg->phraser_mode & MILENA_PHR_IGNORE_INFO)) {
			if (**str=='{' || **str=='[') {
				if (!c) c=*str;
				d=strchr(*str,((**str)=='{')?'}':']');
				if (d) {
					*str=d+1;
					continue;
				}
			}
		}
		if (**str == '-') {
			dialog=2;
			if (mi_spell(cfg,'-',0,NULL)) break;
		}
		else if (good_char(cfg,**str) || mi_spell(cfg,**str,0,NULL)) break;
	}
	if (!**str) return 0;
	d=*str;
	if (c) *str=c;
	for (;*d;d++) if (good_char(cfg,*d) || (spell_empty && mi_spell(cfg,**str,0,NULL))) return dialog;
	return 0;
}


#define IS_EMOTICON eof=0;if (c=phraser_get_emot(cfg,*str,str,&eof)) {\
if (pos) {\
pushbuf(' ');\
}\
pushstr(c);\
nword++;\
if (!**str) break;\
if (!eof) continue;}

static struct milena_dic *milena_find_dic(struct milena *cfg,char *word);

static int phraser_at_end(struct milena *cfg,char *s,int strict)
{
	char word[256],*c;
	if (*s=='-') {
		s++;
		if (!*s) return 1;
		if (isdigit(*s) || lci[(*s) & 255]) return 0;
		return 1;
	}
	while (*s && *s != '\n' && isspace(*s)) s++;
	if (*s=='?') return 2;
	if (!*s) return 1;
	if (!strict && lci[(*s) & 255]) {
		struct milena_dic *md;
		for (c=word;*s;) {
			int znak;
			if (isdigit(*s)) return 0;
			*c++=znak=lci[(*s++) & 255];
			if (!znak) break;
		}
		*c=0;
		md=milena_find_dic(cfg,word);
		if (md && (md->flags & MILDIC_SEP)) return 1;
		return 0;
	}
	if (isdigit(*s) || lci[(*s) & 255]) return 0;
	return 1;
}


static int _milena_GetPhraseWithPunct(struct milena *cfg,
	char **str,char *outbuf,int buflen,int *pmode,int punct,char *some)
{
	int pos=0;
	int nword=0;
	char *bword,*c,*ostr=*str;
	int n,typ;
	int last_num;
	int dash,eof,elang;

	int is_samog(char c)
	{
		char *s;
		c=lci[c & 255];
		if (strchr("aeiouy@",c)) return 1;
		s=cfg->redef[c & 255];
		if (!s) return 0;
		for (;*s;s++) if (strchr("aeiouy@",*s)) return 1;
		return 0;
	}
	int is_cap(char *s)
	{
		while (*s && isspace(*s)) s++;
		if (!*s || !good_char(cfg,*s)) return 0;
		return uci[(*s) & 255];
	}
	int is_finale(char *s,char *fin)
	{
		for (;*fin;s++,fin++) {
			if (!*s) return 0;
			if (lci[(*s) & 255] != *fin) return 0;
		}
		if (good_char(cfg,*s)) return 0;
		return 1;
	}
	int find_mac(char *sq)
	{
		struct mika_afx *mac;
		for (mac=cfg->macanty;mac;mac=mac->next) {
			char *c=mac->afx,*s=sq;
			//fprintf(stderr,"Conp %s %s\n",c,s);
			for (;*c && *s;c++,s++) {
				if (*c=='*') break;
				if (*c!=lci[(*s) & 255]) break;
			}
			if (*c=='*') return 1;
			if (!*c && (!s || !good_char(cfg,*s))) return 1;

		}
		return 0;
	}

	int is_mac(void)
	{
		char *s=NULL,*c,*a;
		int ns;
		a="mek";
		if (!strncasecmp(*str,"o'",2)) {
			s=(*str)+2;
			a="o";
		}
		else if (!strncasecmp(*str,"mc",2)) {
			s=(*str)+2;
			if (*s=='h') return 0;
		}
		else if (!strncasecmp(*str,"mac",3)) {
			int iq;
			s=(*str)+3;
			for (iq=0;s[iq];iq++) if (!good_char(cfg,s[iq])) break;
			if (iq<3) return 0;
			//if (uci[(*s)&255] && !uci[(*str)[2] & 255]) goto macok;
			/*
			if (*s=='y') {
				if (!good_char(cfg,s[1])) return 0;
				goto macok;
			}
			if (*s=='i') {
				if (is_samog(s[1])) return 0;
				if (!strncasecmp(s+1,"sz",2)) return 0;
				if (strchr("񳿼ck",s[1])) return 0;
				goto macok;
			}
			*/
			if (find_mac(s)) return 0;
		}
macok:		if (!s) return 0;
		ns=0;
		for (c=s;*c;c++) {
			if(isdigit(*c)) return 0;
			if (!good_char(cfg,*c)) break;
			if(is_samog(*c)) ns++;
		}
		if (!ns) return 0;
		if (c<s+3) return 0;
		if (pos) pushbuf(' ');
		pushstr("[n]");
		pushstr(a);
		*str=s;
		return 1;
	}

	int is_url(void)
	{
		char *proto;
		char *port;
		char *path;
		char *domain[16],dip[16];
		int domain_num;
		char *c,*s,*end;
		char *tmp;
		int dots=0,ldot=0;
		int domain_is_ip;
		int i,shdot;
		char tbuf[256];

		// funkcja jest na razie zablokowana
		return 0;


		int nicky(char *cs)
		{
			int ns=0,nl=0;
			for (;*cs;cs++) {
				if (is_samog(*cs)) ns++;
				nl++;
			}
			fprintf(stderr,"NL %d NS %d\n",nl,ns);
			if (nl < 3 || !ns || ns == nl) return 0;
			if (!*cs) return 1;
			while (*cs && isdigit(*cs)) cs++;
			if (cs) return 0;
			return 1;
		}
		int nodot_fin(int ndom)
		{
			static char *fina[]={"org","gov","com","net","info","prv",NULL};
			int i;
			if (ndom == domain_num -2) {
				if (strlen(domain[ndom+1]) == 2) {
					return 1;
				}
				for (i=0;fina[i];i++) if (!strcasecmp(fina[i],domain[ndom+1])) return 1;
				return 0;
			}
			if (ndom == domain_num - 3) {
				if (!strcasecmp(domain[ndom+1],"eu") && !strcasecmp(domain[ndom+2],"org")) return 1;
				if (strlen(domain[ndom+1])<3) return 0;
				if (strlen(domain[ndom+2])>2) return 0;
				for (i=0;fina[i];i++) if (!strcasecmp(fina[i],domain[ndom+1])) return 1;
				return 0;
			}
			return 0;
		}

		void emit_url_p(char *cin)
		{
			int spa=1;
			char last,*p;
			static char *fines[]={"html","htm","shtml","asp","cgi","php3","php4","php",NULL};
			int i;
			while (*cin) {
				if (*cin == '%') {
					if (!spa) pushbuf(' ');
					spa=0;
					pushstr("procent");
					cin++;
					if (*cin && isxdigit(*cin) && cin[1] && isxdigit(cin[1])) {
						last=cin[2];
						cin[2]=0;
						p=cin;
						pushbuf(' ');
						pos=milena_spell(cfg,&p,outbuf,buflen,pos,-1);
						cin=cin+2;
						*cin=last;
					}
					continue;
				}
				if (*cin == '.' && cin[1] != '.') {
					cin++;
					for (i=0;fines[i];i++) {
						if (!strcasecmp(cin,fines[i])) break;
					}
					if (fines[i]) {
						if (!spa) {
							pushbuf(' ');
						}
						spa=0;
						p=fines[i];
						pos=milena_spell(cfg,&p,outbuf,buflen,pos,-1);
						return;
					}
				}
				p=cin;
				pos=milena_spell(cfg,&p,outbuf,buflen,pos,-1);
				return;
			}

		}

                void url_emit(char *cin,int mode)
                {
			char *c,*cval;
			if (!*cin) return;
			if (mode) {
				cval=strchr(cin,'=');
				if (cval) *cval++=0;
				else mode=0;
			}
			emit_url_p(cin);
			if (cval) {
				pushstr(" rwne ");
				emit_url_p(cval);
			}
		}


		//return 0;

		s=*str;
		if (!isascii(*s) || !isalnum(*s)) return 0;
		for (s=*str;*s;s++) {
			if (!isascii(*s) || *s==0x7f) return 0; // nie moe by url
			if (*s=='.') ldot=1;
			else if (ldot) {
				ldot=0;
				if (isalnum(*s)) dots++;
			}
			if (*s<=0x20) break;
		}
		end=s;
		s=*str;
		if (!dots) return 0;
		if (end-s > 255) return 0;
		memcpy(tbuf,s,end-s);
		tbuf[end-s]=0;
		tmp=s=tbuf;
		fprintf(stderr,"%s\n",tbuf);
		proto=port=path=NULL;
		if (isalpha(*s)) {
			s++;
			while (isalpha(*s)) s++;
			if (*s==':') {
				proto=tmp;
				*s++=0;
				if (*s=='/' && s[1] == '/') s+=2;
			}
			else {
				s=tmp;
			}
		}
		domain_num=0;
		domain_is_ip=1;
		for (;;) {
			if (!isalnum(*s)) break;
			if (domain_num == 16) return 0;
			domain[domain_num]=s;
			dip[domain_num]=1;
			while (*s && (isalnum(*s) || *s=='-')) {
				if (!isdigit(*s)) {
					domain_is_ip=0;
					dip[domain_num]=0;
				}
				s++;
			}
			domain_num++;
			if (*s!='.') break;
			*s++=0;
			if (*s && !isalnum(*s)) return 0;
		}
		fprintf(stderr,"Domain num=%d\n",domain_num);
		if (domain_num < 2) return 0;
		if (strlen(domain[domain_num-1])<2) return 0;
		if (!*s) {end=s;goto realize;}
		if (*s==':') {
			/* port? */
			tmp=s;
			s++;
			if (*s && isdigit(*s)) {
				port=s;
				while (*s && isdigit(*s)) s++;
				if (*s && isalpha(*s)) return 0;
				if (!*s) {*tmp=0;end=s;goto realize;}
			}
			else s=tmp;
		}
		if (*s=='/') {
			*s++=0;
			if (!*s) {end=s;goto realize;}
			goto paths;
		}
		tmp=s;
		for (;*s;s++) if (isalnum(*s)) return 0;
		end=tmp;
		goto realize;
paths:		path=s;
		//end=path;
		end=s+strlen(s);
		goto realize;
realize:

		/* if (domain_num == 4 && !port && !path && !proto) { */
		if (domain_is_ip) {
			fprintf(stderr,"To ipadr\n");
			return 0;
		}
		(*str)+=end-tbuf;
		if (pos) {
			pushbuf(' ');
		}
		if (proto) {
			fprintf(stderr,"p1 %d\n",pos);
			pos=milena_spell(cfg,&proto,outbuf,buflen,pos,-1);
			fprintf(stderr,"p2 %d\n",pos);
			pushbuf(' ');
		}
		shdot=0;
		for (i=0;i<domain_num;i++) {
			if (dip[i]) {
				int m1,m2;
				if (shdot) {
					pushstr("kropka ");
				}
				c=domain[i];
				pos=milena_spell(cfg,&c,outbuf,buflen,pos,-1);
				if (i<domain_num-1) {
					shdot=0;
					if (!dip[i+1]) shdot=1;
					else {
						m1=strtol(domain[i],NULL,10);
						m2=strtol(domain[i+1],NULL,10);
						if (m1 && m2) {
							if (m2<10 && !(m1%10)) shdot=1;
							else if (m2<100 && !(m1%100)) shdot=1;
						}
					}
					pushbuf(' ');
				}

				continue;
			}
			// cfg->ext_verbs_no - sprawdzam tryb lektora
			c=domain[i];
			if (shdot) {
				pushstr("kropka ");
			}

			if (!strcasecmp(c,"com")) {
				pushstr("kom");
			}
			else if (!strcasecmp(c,"org") || strlen(c)<3 || !nicky(c)) {
				pos=milena_spell(cfg,&c,outbuf,buflen,pos,-1);
			}
			else {
				for (;;) {
                                    pos=milena_nicky(cfg,&c,outbuf,buflen,pos,-1);
                                    if (*c == '-') {
                                        while (*c == '-') {
                                            pushstr(" mylnik");
                                            c++;
                                        }
                                        pushbuf(' ');
                                    }
                                    else break;
                                }
			}
                        pushbuf(' ');
			shdot=nodot_fin(i)?0:1;

		}
		if (port) {
			pushstr(" port ");
			pos=milena_spell(cfg,&port,outbuf,buflen,pos,-1);
		}
		if (!path) return 1;

                int znak='/';int tryb;
                c=strpbrk(path,"/?#");
                for (;;) {
                    if (znak=='#') c=NULL;
                    else {
                        c=strpbrk(path,(znak=='/')?"/?#":"&#");
                    }
                    tryb=(znak == '?' || znak=='&')?1:0;
                    pushstr((znak=='/')?" slesz ":(znak == '?')?" pytajnik":(znak == '?')?" and":" hasz");
                    if (c) {
                        znak=*c;
                        *c++=0;
                    }
                    if (*path) {
                        pushbuf(' ');
                        url_emit(path,tryb);
                    }
		    if (!c) break;
		    path=c;


                }




		return 1;
	}

	void pushy(int z)
	{
		z=lci[z & 255];
		if (cfg->redef[z & 255]) {
			pushstr(cfg->redef[z & 255]);
		} else {
			pushbuf(z);
		}

	}
	int is_apostr(void)
	{
		char *s=*str,lc,*c;
		int rc=1;
		int i;
		int sag=0;
		for (;*s;s++) {
			if (!good_char(cfg,*s)) break;
			lc=lci[(*s) & 255];
			if (is_samog(lc)) sag=1;
		}
		if (*s!='\'') return 0;
		if (s == (*str)+1) {
			char *s1;
			int nextap=0;
			//if (lc != 'l' && lc !='d' && lc !='L' && lc !='D') return 0;
			s++;
			for (s1=s;*s1;s1++) {
				if ((*s1)=='\'') {
					nextap=1;
					break;
				}
				if (!good_char(cfg,*s1)) break;
				if (isdigit(*s1)) return 0;
				lc=lci[(*s1) & 255];
				if (is_samog(lc)) sag=1;
			}
			if (s1==s) return 0;
			if (!sag) return 0;
			if (pos) {
				pushbuf(' ');
			}
			pushy(**str);
			(*str)++;
			(*str)++;
			pushbuf('&');
			if (nextap) {
				for (i=0;aps[i];i++) if (is_finale(s1+1,aps[i])) break;
				if (!aps[i]) nextap=0;
			}

			if (!nextap) {
				for (;**str;(*str)++) {
					if (!good_char(cfg,**str)) break;
					pushy(**str);
				}
				return 1;
			}
			s=s1;
			goto omit_partest;

		}
		if (!sag) return 0;
		//if (lc !='e' && lc!='y') return 0;
		if (lc == 'n' && s[1]=='t' && !good_char(cfg,s[2])) {
			if (pos) {
				pushbuf(' ');
			}
			for (;*str < s;(*str)++) {
				pushy(**str);
				if ((*str) == s-2 && !is_samog(**str)) {
					pushbuf('Y');
				}
			}
			pushbuf('&');
			goto push_afton;
		}
		for (i=0;aps[i];i++) if (is_finale(s+1,aps[i])) break;
		if (!aps[i]) return 0;
		if (pos) {
			pushbuf(' ');
		}
omit_partest:
#if 0
		for (;*str <s ; (*str)++) {
			pushbuf(**str);
		}
		pushbuf('&');
#else
		for (;*str < s;(*str)++) {
			if (strchr("aesio",aps[i][0]) && (*str)==s-4 && !strncasecmp((*str),"aine",4)) {
				pushstr("ejn");
				(*str)+=4;
				break;
			}
			if (aps[i][0]=='i' && (*str)==s-2 && !strncasecmp((*str),"te",2)) {
				pushstr("c&");
				(*str)+=2;
				break;
			}
			if (strchr("aeio",aps[i][0]) && (*str)==s-3 && !strncasecmp(*str,"ice",3)) {
				pushstr("ajs&");
				(*str)+=3;
				break;
			}
			if (strchr("aeio",aps[i][0]) && (*str)==s-2 && !strncasecmp(*str,"ce",2)) {
				pushstr("s&");
				(*str)+=2;
				break;
			}
			if (strchr("aeio",aps[i][0]) && (*str)==s-4 && !strncasecmp(*str,"ance",4)) {
				pushstr("ens&");
				(*str)+=4;
				break;
			}
			if (aps[i][0]=='s'  && (*str)==s-2 && !strncasecmp(*str,"ce",2)) {
				pushstr("s@");
				(*str)+=2;
				break;
			}
			if (aps[i][0]=='s'  && (*str)==s-2 && !strncasecmp(*str,"ie",2)) {
				pushstr("i");
				(*str)+=2;
				break;
			}
			if (aps[i][0]=='e' && (*str)==s-3 && !strncasecmp(*str,"thy",3)) {
				pushstr("thy&");
				(*str)+=3;
				break;
			}
			if (aps[i][0]=='s'  && (*str)==s-4 && !strncasecmp(*str,"ance",4)) {
				pushstr("ens@");
				(*str)+=4;
				break;
			}
			if (*str == s-1) {
				if (aps[i][0] == 's') {
					pushbuf(**str);
					if (lc=='s' || lc=='z') {
						pushstr("~'i");
					}
				}
				else if (lc=='y' && !strchr("yo",aps[i][0])) {
					pushbuf('&');
					if (aps[i][0] == 'e') {
						pushbuf('i');
					}
					else {
						pushbuf('j');
					}
				}
				else if (**str!='e') {
					pushy(**str);
					pushbuf('&');
				}
				else pushbuf('&');
			}
			else {
				pushy(**str);
			}
		}
#endif
push_afton:
		(*str)++;
		while((**str) && good_char(cfg,**str)) {
			char z=*(*str)++;
			pushy(z);
		}
		return rc;
	}
	int is_real_number(char *s)
	{
		if (!isdigit(*s)) return 0;
		while (*s && isdigit(*s)) s++;
		if (!*s) return 1;
		if (good_char(cfg,*s)) return 0;
		if (*s !='.' && *s!=',') return 1;
		s++;
		if (!*s) return 1;
		if (!good_char(cfg,*s)) return 1;
		if (!isdigit(*s)) return 0;
		while (*s && isdigit(*s)) s++;
		if (!*s) return 1;
		if (good_char(cfg,*s)) return 0;
		if (*s !='.' && *s!=',') return 1;
		s++;
		if (!*s) return 1;
		return !good_char(cfg,*s);
	}


	int is_word(char *s)
	{
		int ns=0;
		int nl=0;
		int nd=0;
		int nw=0;
		int nu=0;
		int nas=0;
		for (;;) {
			if (!(cfg->phraser_mode & MILENA_PHR_IGNORE_TILDE) && *s=='~' && strchr(ESCHAR,s[1])) {
				s+=2;
				continue;
			}
			break;
		}
		if (!*s) return 0;
		if (!good_char(cfg,*s)) return 0;
		for (;*s;s++) {
			if (!(cfg->phraser_mode & MILENA_PHR_IGNORE_TILDE) && *s=='~') {
				if (strchr(ESCHAR,s[1])) {
					nas=-1;
					s++;
					continue;
				}
				break;
			}
			if (!good_char(cfg,*s)) break;
			if (isdigit(*s)) {
				nd++;
				continue;
			}
			if (nd /*|| ((*s) & 0x80)*/) nas=-1;else nas++;
			nl++;
			if (is_samog(*s)) {
				ns++;
				if (strchr("",*s)) nu++;
			}
			else if (strchr("WwZz",*s)) nw++;
		}
		if (nd) {
			if (!nl) return MLWORD_NUMBER;
			if (nas>3 && ns>0 && ns < nas) return MLWORD_NICKY;
			return MLWORD_UNSPEAK;
		}
		if (nl==1 && nu==1) return MLWORD_UNSPEAK;
		if (ns || (nl==1 && nw==1) ) return MLWORD_SPEAKABLE;
		return MLWORD_UNSPEAK;
	}

	int is_entity()
	{
		char *sc=*str;int base=10;
		int wc;
		char ebuf[256];
		/* tu jaki esc powinien by... */
		if (*sc++!='&') return 0;
		//if (*sc++!='#') return 0;
		if (*sc++!='\x1b') return 0;
		if (*sc=='x' || *sc=='X') {
			sc++;
			base=16;
			if (!isxdigit(*sc)) return 0;
		}
		else {
			if (!isdigit(*sc)) return 0;
		}
		wc=strtol(sc,&sc,base);
		if (*sc++!=';') return 0;
		*str=sc;
		pushbuf(' ');
		milena_wchar(cfg,wc,ebuf,256,NULL);
		pushstr(ebuf);
		pushbuf(' ');
		last_num=-1;
		nword++;
		return 1;
	}

	elang=0;

	for (;**str;) {
		IS_EMOTICON
		if (eof) break;
		if ((!(cfg->phraser_mode & MILENA_PHR_IGNORE_INFO) &&
			(**str=='[' || **str=='{')) ||
			is_word(*str) || mi_spell(cfg,**str,punct,some) ||
			((**str)=='-' && is_word((*str)+1)==MLWORD_NUMBER)) break;
		(*str)++;
	}
	if (eof) goto final_part;
	bword=NULL;
	last_num=-1;
	for (;;) {
		while (**str && isspace(**str)) (*str)++;
		if (!**str) break;
		if (!(cfg->phraser_mode && MILENA_PHR_IGNORE_TILDE)) {
			if (!strncmp(*str,"~'",2) || !strncmp(*str,"~+",2)) {
				(*str) += 2;
				continue;
			}
		}
		//is_entity tylko w trybie screenreadera!
		if (cfg->charnms) {
			if (is_entity()) continue;
		}
		IS_EMOTICON
		if (eof) break;
		if (**str=='-') {
			if (is_real_number((*str)+1)) {
				(*str)++;
				if (pos) {
					pushbuf(' ');
				}
				pushstr("minus ");
				pos=milena_digitize(str,outbuf,buflen,pos,0,&last_num,cfg);
				nword++;
				goto floop;
			}
		}
		if (!(cfg->phraser_mode & MILENA_PHR_IGNORE_INFO)) {
			if (**str=='[') {
				char *cs=strchr((*str),']');
				if (!cs) {
					*str+=strlen(*str);
					break;
				}
				if (!bword) bword=*str;
				*str=cs+1;
				continue;
			}
			if (**str=='{') {
				char *cs=strchr((*str),'}');
				if (!cs) {
					*str+=strlen(*str);
					break;
				}
				if (!bword) bword=*str;
				*str=cs+1;
				continue;
			}
		}
		if (last_num >= 0 && !bword) {
			char *c=milena_get_unit(cfg,str,last_num,NULL,0);
			if (c) {
				if (pos) pushbuf(' ');
				nword++;
				pushstr(c);
				last_num=-1;
				goto floop;
			}
		}
		dash=0;
		n=milena_recognize(cfg,str,outbuf,buflen,pos);
		if (n>=0) {
			pos=n;
			nword++;
			goto floop;
		}
		if (is_url()) {
			last_num=-1;
			nword++;
			goto aloop;
		}
		n=is_word(*str);

		if (!n) {
			char *c=mi_spell(cfg,**str,punct,some);
			if (c && (!nword || !strchr(",.!?;:",**str))) {
				if (pos) pushbuf(' ');
				pushstr(c);
				nword++;
				(*str)++;
				last_num=-1;
				continue;
			}
		}
		if (n) {
			struct milena_udic *ud;
			nword++;
			if (!bword) {
				for (ud=cfg->udic;ud;ud=ud->next) {
					int nphr;
					struct phraser_dic_rc phra[16];
					char *cs;
					if (ud->flags & MILDIC_ATSTART) {
						if(pos) continue;
					}
					cs=*str;
					if (!phraser_CompareDic(cfg,ud->word,str,phra,&nphr)) continue;
					if (ud->flags & MILDIC_ATQUE) {
						if (phraser_at_end(cfg,*str,0)!=2) {
							*str=cs;
							continue;
						}
					}

					if (ud->flags & MILDIC_BEFORECAP) {
						if (!is_cap(*str)) {
							*str=cs;
							continue;
						}
					}
					if (ud->flags & MILDIC_ATEND) {
						if (!phraser_at_end(cfg,*str,0)) {
							*str=cs;
							continue;
						}
					}
					if (ud -> flags & MILDIC_SPELL) {
						*str=cs;
						n=MLWORD_UNSPEAK;
						break;
					}
					if (pos) {
						pushbuf(' ');
						if (uci[(*cs) & 255]) {
							pushstr("[F]");
						}
					}
					pos=phraser_OutputDic(cfg,ud,pos,outbuf,buflen,phra,nphr);
					last_num=-1;
					goto floop;
				}

				if (n==MLWORD_SPEAKABLE) {
					int ism=0;
					if (is_mac()) {
						elang=2;
						ism=1;
					}
					if (ism || is_apostr()) {
						last_num=-1;
						goto floop;
					}
				}
				// dla francuskich d'cos i l'cos
				//if ((**str=='d' || **str=='l' || **str=='D' || **str=='L') && (*str)[1]=='\'') {
					if (is_apostr()) {
						last_num=-1;
						goto floop;
					}
				//}
			}
			if (pos) {
				pushbuf(' ');
				if (!bword && n==MLWORD_SPEAKABLE && uci[(**str) & 255]) {
					pushstr("[F]");
				}
				if (!bword && n==MLWORD_SPEAKABLE && elang==2) {
					pushstr("[Len]");
				}

			}
			if (bword) {
				if (n==MLWORD_SPEAKABLE)
					for(;bword<*str;bword++) pushbuf(*bword);
				bword=NULL;
			}
			if (n==MLWORD_SPEAKABLE) {
				for (;**str;(*str)++) {
					char cx,*s;
					if (**str=='~' && !(cfg->phraser_mode & MILENA_PHR_IGNORE_TILDE)) {
						if (strchr(ESCHAR,(*str)[1])) {
							(*str)++;
							pushbuf('~');
							pushbuf(**str);
							continue;
						}
						break;
					}
					cx=lci[(**str) & 255];
					if (!cx || cx=='@') break;
					s=cfg->redef[cx & 255];
					if (s) {
						for (;*s;s++) {
							pushbuf(*s);
						}
						continue;
					}
					pushbuf(cx);
				}
				last_num=-1;
			}
			else if (n==MLWORD_NUMBER) {
				pos=milena_digitize(str,outbuf,buflen,pos,0,&last_num,cfg);
			}
			else if (n==MLWORD_NICKY) {
				pos=milena_nicky(cfg,str,outbuf,buflen,pos,0);
			}
			else {
				pos=milena_spell(cfg,str,outbuf,buflen,pos,last_num);
				last_num=-1;
			}
			elang=0;
floop:			if (**str=='-' && is_word((*str)+1)) dash=1;
		}
aloop:
		if (!dash) {
			IS_EMOTICON
			if (eof) goto final_part;
			if (strchr("()!-,.?:;",**str)) break;
		}
		if (!lci[(**str) & 255] && !mi_spell(cfg,**str,punct,some)) (*str)++;
	}
final_part:
	if (!nword) return -1;
	typ=eof;
	for (;**str;(*str)++) {
		char *c;
		if (isspace(**str)) continue;
		IS_EMOTICON
		if (eof) typ |= eof;
		if (!strchr("()!-,.?:;",**str)) break;
		if (**str=='.') {
			if ((*str)[1]=='.' && (*str)[2]=='.') {
				(*str)+=2;
				if (!typ) typ|=64;
				c=mi_spell(cfg,0x80,punct,some);
				if (c) {
					if (pos) pushbuf(' ');
					pushstr(c);
				}
			}
			else {
				if (c=mi_spell(cfg,'.',punct,some)) {
					if (pos) pushbuf(' ');
					pushstr(c);
				}
				typ |= 1;
			}
			continue;
		}
		c=mi_spell(cfg,**str,punct,some);
		if (c) {
			if (pos) pushbuf(' ');
			pushstr(c);
		}

		if (**str=='!') typ |=8;
		else if (**str=='?') typ |=4;
		else if (**str==':') typ |=16;
		else if (**str==';') typ |=16;
		else if (**str=='-') typ |=32;
		else typ |=2;
	}
	if (!typ) *pmode=0;
	else {
		if (!(typ & 31)) {
			if (typ == 64) {
				if (!**str) *pmode=4;
				else *pmode=7;
			}
			else if (typ==32) *pmode=8+1;
			else *pmode=8+7;
		}
		else if (typ & 4) {*pmode=2;if (typ & 32) (*pmode)+=8;}
		else if (typ & 8) {*pmode=3;if (typ & 32) (*pmode)+=8;}
		else if (typ & 1) *pmode=0;
		else if (typ & 16) *pmode=5;
		else *pmode=1;
	}
	if (!**str) *pmode |= 16;
	if (pos<buflen) {
		outbuf[pos]=0;
		return 0;
	}
	*str=ostr;
	return pos+1;
}

int milena_GetPhraseWithPunct(struct milena *cfg,
	char **str,char *outbuf,int buflen,int *pmode,int punct,char *some)
{
	return _milena_GetPhraseWithPunct(cfg,str,outbuf,buflen,pmode,punct,some);
}
int milena_GetPhrase(struct milena *cfg,
	char **str,char *outbuf,int buflen,int *pmode)
{
	return _milena_GetPhraseWithPunct(cfg,str,outbuf,buflen,pmode,0,NULL);
}

